from hwcomponents import EnergyAreaModel, actionDynamicEnergy
from hwcomponents.scaling import *


class LibraryEstimatorClassBase(EnergyAreaModel):
    @actionDynamicEnergy
    def write(self) -> float:
        return 0

    @actionDynamicEnergy
    def read(self) -> float:
        return 0



"""
@inproceedings{10.1145/3579371.3589062,
author = {Andrulis, Tanner and Emer, Joel S. and Sze, Vivienne},
title = {RAELLA: Reforming the Arithmetic for Efficient, Low-Resolution, and Low-Loss Analog PIM: No Retraining Required!},
year = {2023},
isbn = {9798400700958},
publisher = {Association for Computing Machinery},
address = {New York, NY, USA},
url = {https://doi.org/10.1145/3579371.3589062},
doi = {10.1145/3579371.3589062},
abstract = {Processing-In-Memory (PIM) accelerators have the potential to efficiently run Deep Neural Network (DNN) inference by reducing costly data movement and by using resistive RAM (ReRAM) for efficient analog compute. Unfortunately, overall PIM accelerator efficiency is limited by energy-intensive analog-to-digital converters (ADCs). Furthermore, existing accelerators that reduce ADC cost do so by changing DNN weights or by using low-resolution ADCs that reduce output fidelity. These strategies harm DNN accuracy and/or require costly DNN retraining to compensate.To address these issues, we propose the RAELLA architecture. RAELLA adapts the architecture to each DNN; it lowers the resolution of computed analog values by encoding weights to produce near-zero analog values, adaptively slicing weights for each DNN layer, and dynamically slicing inputs through speculation and recovery. Low-resolution analog values allow RAELLA to both use efficient low-resolution ADCs and maintain accuracy without retraining, all while computing with fewer ADC converts.Compared to other low-accuracy-loss PIM accelerators, RAELLA increases energy efficiency by up to 4.9\texttimes{} and throughput by up to 3.3\texttimes{}. Compared to PIM accelerators that cause accuracy loss and retrain DNNs to recover, RAELLA achieves similar efficiency and throughput without expensive DNN retraining.},
booktitle = {Proceedings of the 50th Annual International Symposium on Computer Architecture},
articleno = {27},
numpages = {16},
keywords = {processing in memory, compute in memory, analog, neural networks, accelerator, architecture, slicing, ADC, ReRAM},
location = {Orlando, FL, USA},
series = {ISCA '23}
}
"""

from hwcomponents_library.base import LibraryEstimatorClassBase
from hwcomponents.scaling import *
from hwcomponents import actionDynamicEnergy


# Original CSV contents:
# tech_node,global_cycle_period,energy,area,n_instances,action
# 40nm,1e-9,0.25,0,1,multiply|read,
# 40nm,1e-9,0,0,1,update|leak|write,
# # Assuming multiplication energy scales linearly with input, weight, and output energy
# # Efficient processing of DNNs (Sze, 2020): 8b*8b->16b multiply 0.2pJ
# # 16b * 8b -> 8b multiply: 0.2 pJ
# # We do this at the L2 (large) tile level, so area will be negligible
class RaellaQuantMultiplier(LibraryEstimatorClassBase):
    """
    The quantization & multipliler from the RAELLA paper. This unit will multiply a
    partial sum value by a quantization scale to apply linear quntization.

    Parameters
    ----------
    tech_node: float
        Technology node in meters.
    """

    component_name = "raella_quant_multiplier"
    priority = 0.9

    def __init__(self, tech_node: float, resolution: int=16):
        super().__init__(leak_power=0.0, area=0.0e-12)
        self.tech_node: float = self.scale(
            "tech_node",
            tech_node,
            40e-9,
            tech_node_energy,
            tech_node_area,
            tech_node_leak,
        )
        self.resolution: int = self.scale(
            "resolution",
            resolution,
            16,
            linear,
            linear,
            linear
        )

    @actionDynamicEnergy
    def multiply(self) -> float:
        """
        Returns the energy consumed by a multiply operation in Joules.

        Returns
        -------
        float
            The energy consumed by a multiply operation in Joules.
        """
        return 0.25e-12

    @actionDynamicEnergy
    def read(self) -> float:
        """
        Returns the energy consumed by a multiply operation in Joules.

        Returns
        -------
        float
            The energy consumed by a multiply operation in Joules.
        """
        return 0.25e-12



from hwcomponents_library.library.aladdin import *
from hwcomponents_library.library.atomlayer import *
from hwcomponents_library.library.brahms import *
from hwcomponents_library.library.dummy import *
from hwcomponents_library.library.forms import *
from hwcomponents_library.library.isaac import *
from hwcomponents_library.library.jia import *
from hwcomponents_library.library.misc import *
from hwcomponents_library.library.newton import *
from hwcomponents_library.library.raella import *
from hwcomponents_library.library.timely import *
from hwcomponents_library.library.wan import *



"""
@article{wan2022compute,
  title={A compute-in-memory chip based on resistive random-access memory},
  author={Wan, Weier and Kubendran, Rajkumar and Schaefer, Clemens and Eryilmaz, Sukru Burc and Zhang, Wenqiang and Wu, Dabin and Deiss, Stephen and Raina, Priyanka and Qian, He and Gao, Bin and others},
  journal={Nature},
  volume={608},
  number={7923},
  pages={504--512},
  year={2022},
  publisher={Nature Publishing Group UK London}
}
"""

from hwcomponents_library.base import LibraryEstimatorClassBase
from hwcomponents.scaling import *
from hwcomponents import actionDynamicEnergy


# Original CSV contents:
# tech_node,n_repeats,global_cycle_period,resolution, voltage,energy,area,  action
# 130nm,     1,      1e-6,                8,          1.8,    0.1,   170,  read
# 130nm,     1,      1e-6,                8,          1.8,    0.1,   170,  write|update
# 130nm,     1,      1e-6,                8,          1.8,    0.1,   170,  leak
class WanShiftAdd(LibraryEstimatorClassBase):
    """
    The shift-and-add unit from the Wan et al. Nature 2022 paper. This unit will sum and
    accumulate values in a register, while also shifting the register contents to accept
    various power-of-two scaling factors for the summed values.

    Parameters
    ----------
    tech_node: float
        Technology node in meters.
    n_repeats: int
        Number of times to repeat the shift-and-add operation.
    resolution: int
        Resolution of the shift-and-add unit in bits. This is the number of bits of each
        input value that is added to the register.
    voltage: float
        Voltage of the shift-and-add unit in volts.
    """

    component_name = "wan_shift_add"
    priority = 0.9

    def __init__(
        self,
        tech_node: float,
        n_repeats: int = 1,
        resolution: int = 8,
        voltage: float = 1.8,
    ):
        super().__init__(leak_power=1.00e-7, area=170.0e-12)
        self.tech_node: float = self.scale(
            "tech_node",
            tech_node,
            130e-9,
            tech_node_energy,
            tech_node_area,
            tech_node_leak,
        )
        self.n_repeats: int = self.scale(
            "n_repeats", n_repeats, 1, linear, linear, linear
        )
        self.resolution: int = self.scale(
            "resolution", resolution, 8, pow_base(2), pow_base(2), pow_base(2)
        )
        self.voltage: float = self.scale("voltage", voltage, 1.8, noscale, quadratic, 1)

    @actionDynamicEnergy
    def read(self) -> float:
        """
        Returns zero Joules.

        Returns
        -------
        float
            Zero Joules.
        """
        return 0

    @actionDynamicEnergy
    def write(self) -> float:
        """
        Returns the energy used to perform a shift-and-add operation in Joules.

        Returns
        -------
        float
            The energy used to perform a shift-and-add operation in Joules.
        """
        return self.shift_and_add()

    @actionDynamicEnergy
    def shift_and_add(self) -> float:
        """
        Returns the energy used to perform a shift-and-add operation in Joules.

        Returns
        -------
        float
            The energy used to perform a shift-and-add operation in Joules.
        """
        return 0.1e-12


# Original CSV contents:
# tech_node,n_repeats,global_cycle_period,voltage,energy,area,  action
# 130nm,     1,      1e-6,                1.8,    0.3,  400,   read
# 130nm,     1,      1e-6,                1.8,    0,     400,   leak
# 130nm,     1,      1e-6,                1.8,    0,     400,   write|update
class WanVariablePrecisionADC(LibraryEstimatorClassBase):
    """
    The variable precision ADC from the Wan et al. Nature 2022 paper. This unit will
    convert a voltage to a digital value.

    Parameters
    ----------
    tech_node: float
        Technology node in meters.
    n_repeats: int
        Number of times to repeat the ADC operation.
    voltage: float
        Voltage of the ADC unit in volts.
    """

    component_name = "wan_variable_precision_adc"
    priority = 0.9

    def __init__(self, tech_node: float, n_repeats: int = 1, voltage: float = 1.8):
        super().__init__(leak_power=0.0, area=400.0e-12)
        self.tech_node: float = self.scale(
            "tech_node",
            tech_node,
            130e-9,
            tech_node_energy,
            tech_node_area,
            tech_node_leak,
        )
        self.n_repeats: int = self.scale(
            "n_repeats", n_repeats, 1, linear, linear, linear
        )
        self.voltage: float = self.scale("voltage", voltage, 1.8, noscale, quadratic, 1)

    @actionDynamicEnergy
    def convert(self) -> float:
        """
        Returns the energy used to convert a voltage to a digital value in Joules.

        Returns
        -------
        float
            The energy used to convert a voltage to a digital value in Joules.
        """
        return 0.3e-12

    @actionDynamicEnergy
    def read(self) -> float:
        """
        Returns the energy used to convert a voltage to a digital value in Joules.

        Returns
        -------
        float
            The energy used to convert a voltage to a digital value in Joules.
        """
        return self.convert()


# Original CSV contents:
# tech_node,global_cycle_period,voltage,energy,area,  action
# 130nm,     1e-6,                1.8,    1.2,   350,  read
# 130nm,     1e-6,                1.8,    0,     350,  leak
# 130nm,     1e-6,                1.8,    0,     350,  write|update
class WanAnalogSample(LibraryEstimatorClassBase):
    """
    The analog sample unit from the Wan et al. Nature 2022 paper. This unit will sample
    an analog charge and add it to an analog integrator.

    Parameters
    ----------
    tech_node: float
        Technology node in meters.
    voltage: float
        Voltage of the analog sample unit in volts.
    """

    component_name = "wan_analog_sample"
    priority = 0.9

    def __init__(self, tech_node: float, voltage: float = 1.8):
        super().__init__(leak_power=0.0, area=350.0e-12)
        self.tech_node: float = self.scale(
            "tech_node",
            tech_node,
            130e-9,
            tech_node_energy,
            tech_node_area,
            tech_node_leak,
        )
        self.voltage: float = self.scale("voltage", voltage, 1.8, noscale, quadratic, 1)

    @actionDynamicEnergy
    def read(self) -> float:
        """
        Returns the energy used to sample an analog voltage in Joules.

        Returns
        -------
        float
            The energy used to sample an analog voltage in Joules.
        """
        return self.sample()

    @actionDynamicEnergy
    def sample(self) -> float:
        """
        Returns the energy used to sample an analog voltage in Joules.

        Returns
        -------
        float
            The energy used to sample an analog voltage in Joules.
        """
        return 1.2e-12


# Original CSV contents:
# tech_node,n_repeats,global_cycle_period,voltage,energy,area, action
# 130nm,     1,      1e-6,                 1.8,   0.25,  350,  read
# 130nm,     1,      1e-6,                1.8,    0,     350,  leak
# 130nm,     1,      1e-6,                1.8,    0,     350,  write|update
class WanAnalogIntegrator(LibraryEstimatorClassBase):
    """
    The analog integrator unit from the Wan et al. Nature 2022 paper. This unit will
    integrate charge over multiple samples.

    Parameters
    ----------
    tech_node: float
        Technology node in meters.
    n_repeats: int
        Number of times to repeat the analog integration operation.
    voltage: float
        Voltage of the analog integrator unit in volts.
    """

    component_name = "wan_analog_integrator"
    priority = 0.9

    def __init__(self, tech_node: float, n_repeats: int = 1, voltage: float = 1.8):
        super().__init__(leak_power=0.0, area=350.0e-12)
        self.tech_node: float = self.scale(
            "tech_node",
            tech_node,
            130e-9,
            tech_node_energy,
            tech_node_area,
            tech_node_leak,
        )
        self.n_repeats: int = self.scale(
            "n_repeats", n_repeats, 1, linear, linear, linear
        )
        self.voltage: float = self.scale("voltage", voltage, 1.8, noscale, quadratic, 1)

    @actionDynamicEnergy
    def integrate(self) -> float:
        """
        Returns the energy used to integrate charge for a sample in Joules.

        Returns
        -------
        float
            The energy used to integrate charge for a sample in Joules.
        """
        return 0.25e-12

    @actionDynamicEnergy
    def read(self) -> float:
        """
        Returns the energy used to integrate charge for a sample in Joules.

        Returns
        -------
        float
            The energy used to integrate charge for a sample in Joules.
        """
        return self.integrate()



"""
@INPROCEEDINGS{8465832,
  author={Qiao, Ximing and Cao, Xiong and Yang, Huanrui and Song, Linghao and Li, Hai},
  booktitle={2018 55th ACM/ESDA/IEEE Design Automation Conference (DAC)},
  title={AtomLayer: A Universal ReRAM-Based CNN Accelerator with Atomic Layer Computation},
  year={2018},
  volume={},
  number={},
  pages={1-6},
  doi={10.1109/DAC.2018.8465832}}
"""

from hwcomponents_library.base import LibraryEstimatorClassBase
from hwcomponents.scaling import *
from hwcomponents import actionDynamicEnergy
from .isaac import IsaacADC
from .isaac import IsaacDAC
from .isaac import IsaacEDRAM
from .isaac import IsaacEDRAMBus
from .isaac import IsaacRouter
from .isaac import IsaacShiftAdd


# Original CSV contents:
# tech_node,global_cycle_period,width|datawidth,depth,energy,area,action
# 32nm,1e-9,16,128,0.083,1620,read|write,energy in pJ;  area in um^2;
# 32nm,1e-9,16,128,0,1620,update|leak
# # 1 read, 1 write per DAC activation
# # Reported power = 0.39 / 4 x DAC
# # 0.166015625 / 2 = 0.083
class AtomlayerRegisterLadder(LibraryEstimatorClassBase):
    """
    A register ladder from the AtomLayer paper. Is a series of registers that shift
    stored values along themselves.

    Parameters
    ----------
    tech_node : str
        The technology node in meters.
    width : int, optional
        The width of the register ladder in bits. This is the bits in each register.
        Total size = width * depth.
    depth : int, optional
        The number of entries in the register ladder, each with `width` bits. Total size
        = width * depth.
    """

    component_name = "atomlayer_register_ladder"
    priority = 0.9

    def __init__(self, tech_node: float, width: int = 16, depth: int = 128):
        super().__init__(leak_power=0.0, area=1620.0e-12)
        self.tech_node: float = self.scale(
            "tech_node",
            tech_node,
            32e-9,
            tech_node_energy,
            tech_node_area,
            tech_node_leak,
        )
        self.width: int = self.scale("width", width, 16, linear, linear, linear)
        self.depth: int = self.scale(
            "depth", depth, 128, linear, cacti_depth_energy, cacti_depth_energy
        )

    @actionDynamicEnergy(bits_per_action="width")
    def read(self) -> float:
        """
        Returns the energy for one read operation in Joules.

        Parameters
        ----------
        bits_per_action : int
            The number of bits that are read.

        Returns
        -------
        float
            The energy for one read operation in Joules.
        """
        return 0.083e-12

    @actionDynamicEnergy(bits_per_action="width")
    def write(self) -> float:
        """
        Returns the energy for one write operation in Joules.

        Parameters
        ----------
        bits_per_action : int
            The number of bits that are written.

        Returns
        -------
        float
            The energy for one write operation in Joules.
        """
        return 0.083e-12


# Original CSV contents:
# tech_node,global_cycle_period,width|datawidth,depth,energy,area,action
# 32nm,1e-9,16,128,6.46,2100,read, energy in pJ;  area in um^2;
# 32nm,1e-9,16,128,0,2100,write|update|leak
# # Power calculation for input buffers:
# # Power * Time / (Reads+Writes) = Energy per read/write
# # (1.24e-3 W) power * (16 * 100e-9s time/MAC / 1.2) / (128+128 reads+writes)
# # (1.24e-3) * (16 * 100e-9 / 1.2) / (128+128) * 1e12
# # Now for the transfers calculation, we also mark write energy = 0 so we don't
# # double charge for writes with the actual buffers. Only charge for reads when
# # another
# # buffer reads from the inter-buffer transfer network.
class AtomlayerInputBufferTransfers(LibraryEstimatorClassBase):
    """
    This component measures transfer energy between input buffers in the AtomLayer
    paper.

    Parameters
    ----------
    tech_node : str
        The technology node in meters.
    width : int, optional
        The width of the read/write port of each input buffer in bits. Total size =
        width * depth.
    depth : int, optional
        The number of entries in each input buffer, each with `width` bits. Total size =
        width * depth.
    """

    component_name = "atomlayer_input_buffer_transfers"
    priority = 0.9

    def __init__(self, tech_node: float, width: int = 16, depth: int = 128):
        super().__init__(leak_power=0.0, area=2100.0e-12)
        self.tech_node: float = self.scale(
            "tech_node",
            tech_node,
            32e-9,
            tech_node_energy,
            tech_node_area,
            tech_node_leak,
        )
        self.width: int = self.scale("width", width, 16, linear, linear, linear)
        self.depth: int = self.scale(
            "depth", depth, 128, linear, cacti_depth_energy, cacti_depth_energy
        )

    @actionDynamicEnergy(bits_per_action="width")
    def read(self) -> float:
        """
        Returns the energy for one read operation in Joules.

        Parameters
        ----------
        bits_per_action : int
            The number of bits that are transferred.

        Returns
        -------
        float
            The energy for one transfer operation in Joules.
        """
        return 6.46e-12

    @actionDynamicEnergy(bits_per_action="width")
    def transfer(self) -> float:
        """
        Returns the energy for one transfer operation in Joules.

        Parameters
        ----------
        bits_per_action : int
            The number of bits that are transferred.

        Returns
        -------
        float
            The energy for one transfer operation in Joules.
        """
        return 6.46e-12


class AtomlayerADC(IsaacADC):
    pass


class AtomlayerDAC(IsaacDAC):
    pass


class AtomlayerRouter(IsaacRouter):
    pass


class AtomlayerEDRAM(IsaacEDRAM):
    pass


class AtomlayerEDRAMBus(IsaacEDRAMBus):
    pass


class AtomlayerShiftAdd(IsaacShiftAdd):
    pass



"""
@ARTICLE{9082159,
  author={Jia, Hongyang and Valavi, Hossein and Tang, Yinqi and Zhang, Jintao and Verma, Naveen},
  journal={IEEE Journal of Solid-State Circuits},
  title={A Programmable Heterogeneous Microprocessor Based on Bit-Scalable In-Memory Computing},
  year={2020},
  volume={55},
  number={9},
  pages={2609-2621},
  doi={10.1109/JSSC.2020.2987714}}
"""

from hwcomponents_library.base import LibraryEstimatorClassBase
from hwcomponents.scaling import *
from hwcomponents import actionDynamicEnergy


# Original CSV contents:
# tech_node,global_cycle_period,resolution,voltage,energy,area,action
# 65nm,      540e-9,              8,         1.2,   2.25,   5000,read
# 65nm,      540e-9,              8,         1.2,   1.2,    5000,leak
# 65nm,      540e-9,              8,         1.2,   0,      5000,write|update
class JiaShiftAdd(LibraryEstimatorClassBase):
    """
    The shift-and-add unit from Jia et al. JSSC 2020. This unit will sum and accumulate
    values in a register, while also shifting the register contents to accept various
    power-of-two scaling factors for the summed values.

    Parameters
    ----------
    tech_node: float
        Technology node in meters.
    resolution: int
        Resolution of the shift-and-add unit in bits. This is the number of bits of each
        input value that is added to the register.
    voltage: float
        Voltage of the shift-and-add unit in volts.
    """

    component_name = "jia_shift_add"
    priority = 0.9

    def __init__(self, tech_node: float, resolution: int = 8, voltage: float = 1.2):
        super().__init__(leak_power=2.22e-6, area=5000.0e-12)
        self.tech_node: float = self.scale(
            "tech_node",
            tech_node,
            65e-9,
            tech_node_energy,
            tech_node_area,
            tech_node_leak,
        )
        self.resolution: int = self.scale(
            "resolution", resolution, 8, pow_base(2), pow_base(2), pow_base(2)
        )
        self.voltage: float = self.scale("voltage", voltage, 1.2, noscale, quadratic, 1)

    @actionDynamicEnergy
    def shift_and_add(self) -> float:
        """
        Returns the energy consumed by a shift+add operation in Joules.

        Returns
        -------
        float
            The energy consumed by a shift+add operation in Joules.
        """

        return 2.25e-12

    @actionDynamicEnergy
    def write(self) -> float:
        """
        Returns the energy consumed by a shift+add operation in Joules.

        Returns
        -------
        float
            The energy consumed by a shift+add operation in Joules.
        """
        return self.shift_and_add()

    @actionDynamicEnergy
    def read(self) -> float:
        """
        Zero Joules to read

        Returns
        -------
        float
            Zero
        """
        return 0.0


# Original CSV contents:
# tech_node,global_cycle_period,rows,resolution,voltage,energy,area,action
# 65nm,      540e-9,              1,   8,         1.2,   0.5,   174, read
# 65nm,      540e-9,              1,   8,         1.2,   0.2,   174, leak
# 65nm,      540e-9,              1,   8,         1.2,   0,     174, write|update
class JiaZeroGate(LibraryEstimatorClassBase):
    """
    The zero gating unit from Jia et al. JSSC 2020. This unit gates analog voltages for
    zero-valued inputs going into the rows of the crossbar array.

    Parameters
    ----------
    tech_node: float
        Technology node in meters.
    rows: int
        Number of rows in the crossbar array, equal to the number of checks done by the
        zero gate.
    resolution: int
        Resolution of each input in bits.
    voltage: float
        Voltage of the zero gating unit in volts.
    """

    component_name = "jia_zero_gate"
    priority = 0.9

    def __init__(
        self, tech_node: float, rows: int = 1, resolution: int = 8, voltage: float = 1.2
    ):
        super().__init__(leak_power=3.70e-7, area=174.0e-12)
        self.tech_node: float = self.scale(
            "tech_node",
            tech_node,
            65e-9,
            tech_node_energy,
            tech_node_area,
            tech_node_leak,
        )
        self.rows: int = self.scale("rows", rows, 1, linear, noscale, noscale)
        self.resolution: int = self.scale(
            "resolution", resolution, 8, pow_base(2), pow_base(2), pow_base(2)
        )
        self.voltage: float = self.scale("voltage", voltage, 1.2, noscale, quadratic, 1)

    @actionDynamicEnergy(bits_per_action="resolution")
    def zero_gate(self) -> float:
        """
        Returns the energy consumed to zero gate & read an input in Joules.

        Parameters
        ----------
        bits_per_action: int
            The number of bits to check for zero.

        Returns
        -------
        float
            The energy consumed to zero gate & read an input in Joules.
        """
        return 0.5e-12

    def read(self) -> float:
        """
        Returns the energy consumed to zero gate & read an input in Joules.

        Parameters
        ----------
        bits_per_action: int
            The number of bits to check for zero.

        Returns
        -------
        float
            The energy consumed to zero gate & read an input in Joules.
        """
        return self.zero_gate()


# Original CSV contents:
# tech_node,global_cycle_period,voltage,energy,area,  action
# 65nm,      540e-9,              1.2,   12,     10535,read
# 65nm,      540e-9,              1.2,   2.4,    10535,leak
# 65nm,      540e-9,              1.2,   0,      10535,write|update
class JiaDatapath(LibraryEstimatorClassBase):
    """
    The datapath in Jia et al. JSSC 2020. This datapath will perform quantization and
    activation functions on accumulated ouputs.

    Parameters
    ----------
    tech_node: float
        Technology node in meters.
    voltage: float
        Voltage of the datapath in volts.

    """

    component_name = "jia_datapath"
    priority = 0.9

    def __init__(self, tech_node: float, voltage: float = 1.2):
        super().__init__(leak_power=4.44e-6, area=10535.0e-12)
        self.tech_node: float = self.scale(
            "tech_node",
            tech_node,
            65e-9,
            tech_node_energy,
            tech_node_area,
            tech_node_leak,
        )
        self.voltage: float = self.scale("voltage", voltage, 1.2, noscale, quadratic, 1)

    @actionDynamicEnergy
    def process(self) -> float:
        """
        Returns the energy consumed by the datapath to quantize and apply activation
        functions on a single input.

        Returns
        -------
        float
            The energy consumed by the datapath to process an input in Joules.
        """
        return 2.4e-12

    @actionDynamicEnergy
    def read(self) -> float:
        """
        Returns the energy consumed by the datapath to quantize and apply activation
        functions on a single input.

        Returns
        -------
        float
            The energy consumed by the datapath to process an input in Joules.
        """
        return self.process()



"""
@ARTICLE{8474954,
  author={Nag, Anirban and Balasubramonian, Rajeev and Srikumar, Vivek and Walker, Ross and Shafiee, Ali and Strachan, John Paul and Muralimanohar, Naveen},
  journal={IEEE Micro},
  title={Newton: Gravitating Towards the Physical Limits of Crossbar Acceleration},
  year={2018},
  volume={38},
  number={5},
  pages={41-49},
  doi={10.1109/MM.2018.053631140}}
"""

from hwcomponents_library.base import LibraryEstimatorClassBase
from hwcomponents.scaling import *
from hwcomponents import actionDynamicEnergy
from .isaac import IsaacADC
from .isaac import IsaacDAC
from .isaac import IsaacEDRAM
from .isaac import IsaacEDRAMBus
from .isaac import IsaacRouter
from .isaac import IsaacShiftAdd


# Original CSV contents:
# tech_node,global_cycle_period,resolution,energy,area,action
# 32nm,1e-9,9,2.58333333333,1500,convert|read
# 32nm,1e-9,9,0,1500,write|update|leak
# # Energy: 3.1*10^-3 W / (1.2*10^9 ADC BW) * 10 ^ 12 J->pJ
# # Newton's adapative ADC resolution table:
# # 9,9,7,5,1,0,9,3
# # 9,9,8,6,2,0,9,4
# # 9,9,9,7,3,1,9,5
# # 9,9,9,8,4,2,9,6
# # 8,9,9,9,5,3,9,7
# # 7,9,9,9,6,4,9,8
# # 6,8,9,9,7,5,9,9
# # 5,7,9,9,8,6,9,9
# # 4,6,9,9,9,7,8,9
# # 3,5,9,9,9,8,7,9
# # 2,4,8,9,9,9,6,9
# # 1,3,7,9,9,9,5,9
# # 0,2,6,8,9,9,4,9
# # 0,1,5,7,9,9,3,9
# # 0,0,4,6,9,9,2,8
# # 0,0,3,5,9,9,1,7
# # Newton assumes a linear scaling: 9-bit ADC uses X/9 power for X-bit convert.
# # Matches with the table above:
# #    Sum of this table is 832. Sum of full-resolution (all table entries = 9)
# #    is 1152. This is a 40% reduction, matching with the reported 40%
# #    ADC power reduction in the paper.
class NewtonADC(LibraryEstimatorClassBase):
    """
    The ADC from the Newton paper. This is a 9-bit ADC that can optionally stop
    quantizing after a certain number of bits.

    Parameters
    ----------
    tech_node: float
        Technology node in meters.
    resolution: int
        Resolution of the ADC in bits.
    """

    component_name = "newton_adc"
    priority = 0.9

    def __init__(self, tech_node: float, resolution: int = 9):
        super().__init__(leak_power=0.0, area=1500.0e-12)
        self.tech_node: float = self.scale(
            "tech_node",
            tech_node,
            32e-9,
            tech_node_energy,
            tech_node_area,
            tech_node_leak,
        )
        self.resolution: int = self.scale(
            "resolution", resolution, 9, pow_base(2), pow_base(2), pow_base(2)
        )

    @actionDynamicEnergy
    def convert(self) -> float:
        """
        Returns the energy consumed by a convert operation in Joules.

        Returns
        -------
        float
            The energy consumed by a convert operation in Joules.
        """
        return 2.58333333333e-12

    @actionDynamicEnergy
    def read(self) -> float:
        """
        Returns the energy consumed by a convert operation in Joules.

        Returns
        -------
        float
            The energy consumed by a convert operation in Joules.
        """
        return 2.58333333333e-12


class NewtonADC(IsaacADC):
    pass


class NewtonDAC(IsaacDAC):
    pass


class NewtonRouter(IsaacRouter):
    pass


class NewtonEDRAM(IsaacEDRAM):
    pass


class NewtonEDRAMBus(IsaacEDRAMBus):
    pass


class NewtonShiftAdd(IsaacShiftAdd):
    pass



"""
@INPROCEEDINGS{7551379,
  author={Shafiee, Ali and Nag, Anirban and Muralimanohar, Naveen and Balasubramonian, Rajeev and Strachan, John Paul and Hu, Miao and Williams, R. Stanley and Srikumar, Vivek},
  booktitle={2016 ACM/IEEE 43rd Annual International Symposium on Computer Architecture (ISCA)},
  title={ISAAC: A Convolutional Neural Network Accelerator with In-Situ Analog Arithmetic in Crossbars},
  year={2016},
  volume={},
  number={},
  pages={14-26},
  doi={10.1109/ISCA.2016.12}}
"""

from hwcomponents_library.base import LibraryEstimatorClassBase
from hwcomponents.scaling import *
from hwcomponents import actionDynamicEnergy


# Original CSV contents:
# tech_node,global_cycle_period,width|datawidth,depth,energy,area,action
# 32nm,1e-9,256,2048,20.45,83000,read|write|update,energy in pJ;  area in um^2;
# 32nm,1e-9,256,2048,0,83000,leak
# # Power * Time / (Reads+Writes) = Energy per read/write
# # (20.7e-3 / 12 W/IMA) power
# # (16384 / ((128*8*10^7*1.2) * 100 / 128)) time for DACs/ADCs to consume entire input buffer
# # (16384 + 2048) * 2 / 256 reads+writes, including IMA<->eDRAM<->network
# # (20.7e-3 / 12) * (16384 / ((128*8*10^7*1.2) * 100 / 128)) / ((16384 + 2048) * 2 / 256) * 1e12
class IsaacEDRAM(LibraryEstimatorClassBase):
    """
    The embedded DRAM from the ISAAC paper.

    Parameters
    ----------
    tech_node: float
        Technology node in meters.
    width: int
        Width of the eDRAM in bits. This is the width of a reads/write port. Total size =
        width * depth.
    depth: int
        Depth of the eDRAM in bits. This is the number of entries in the eDRAM, each
        with `width` bits. Total size = width * depth.
    """
    component_name = "isaac_eDRAM"
    priority = 0.9

    def __init__(self, tech_node: float, width: int = 256, depth: int = 2048):
        super().__init__(leak_power=0.0, area=83000.0e-12)
        self.tech_node: float = self.scale(
            "tech_node",
            tech_node,
            32e-9,
            tech_node_energy,
            tech_node_area,
            tech_node_leak,
        )
        self.width: int = self.scale("width", width, 256, linear, linear, linear)
        self.depth: int = self.scale(
            "depth", depth, 2048, linear, cacti_depth_energy, cacti_depth_energy
        )

    @actionDynamicEnergy(bits_per_action="width")
    def read(self) -> float:
        """
        Returns the energy of one read operation in Joules.

        Returns
        -------
        float
            Energy of one read operation in Joules
        """
        return 20.45e-12

    @actionDynamicEnergy(bits_per_action="width")
    def write(self) -> float:
        """
        Returns the energy of one write operation in Joules.

        Returns
        -------
        float
            Energy of one write operation in Joules
        """
        return 20.45e-12


# Original CSV contents:
# tech_node,global_cycle_period,width|datawidth,energy,area,action
# 65nm,1e-9,128,26,23000000,read|write|update
# 65nm,1e-9,128,0, 23000000,leak
class IsaacChip2ChipLink(LibraryEstimatorClassBase):
    """
    The chip-to-chip link from the ISAAC paper. This connects multiple chips together.

    Parameters
    ----------
    tech_node: float
        Technology node in meters.
    width: int
        Width of the link in bits. This is the width of a read/write port.
    """
    component_name = "isaac_chip2chip_link"
    priority = 0.9

    def __init__(self, tech_node: float, width: int = 128):
        super().__init__(leak_power=0.0, area=23000000.0e-12)
        self.tech_node: float = self.scale(
            "tech_node",
            tech_node,
            65e-9,
            tech_node_energy,
            tech_node_area,
            tech_node_leak,
        )
        self.width: int = self.scale("width", width, 128, linear, linear, linear)

    @actionDynamicEnergy(bits_per_action="width")
    def read(self) -> float:
        """
        Returns the energy of one read operation in Joules.

        Returns
        -------
        float
            Energy of one read operation in Joules
        """
        return 26.0e-12

    @actionDynamicEnergy(bits_per_action="width")
    def write(self) -> float:
        """
        Returns the energy of one write operation in Joules.

        Returns
        -------
        float
            Energy of one write operation in Joules
        """
        return 26.0e-12


# Original CSV contents:
# tech_node,global_cycle_period,width|datawidth,energy,area,action
# 32nm,1e-9,256,20.74,37500,read,energy in pJ;  area in um^2;
# 32nm,1e-9,256,0,37500,leak|update|write
# # To match the paper where ISAAC shares each of these between 4 tiles. Quarter the area
# # relative to isaac_router
# # Assuming router BW = eDRAM BW per tile
# # Power * Time / (Reads+Writes) = Energy per read/write
# # (42e-3 / 4 / 12) power
# # (16384 / ((128*8*10^7*1.2) * 100 / 128)) time for DACs/ADCs to consume entire input buffer
# # (16384 + 2048) / 256 reads+writes
# # (42e-3 / 4 / 12) * (16384 / ((128*8*10^7*1.2) * 100 / 128)) / ((16384 + 2048) / 256) * 1e12
class IsaacRouterSharedByFour(LibraryEstimatorClassBase):
    """
    This is the router from the ISAAC paper. In the paper, it is shared by four tiles,
    so this area is divided by four to match the paper.
    """

    component_name = "isaac_router_shared_by_four"
    priority = 0.9

    def __init__(self, tech_node: float, width: int = 256):
        super().__init__(leak_power=0.0, area=37500.0e-12)
        self.tech_node: float = self.scale(
            "tech_node",
            tech_node,
            32e-9,
            tech_node_energy,
            tech_node_area,
            tech_node_leak,
        )
        self.width: int = self.scale("width", width, 256, linear, linear, linear)

    @actionDynamicEnergy(bits_per_action="width")
    def read(self) -> float:
        """
        Returns the energy to transfer data in Joules.

        Parameters
        ----------
        bits_per_action: int
            The number of bits transferred.

        Returns
        -------
        float
            Energy to transfer data in Joules
        """
        return self.transfer()

    @actionDynamicEnergy(bits_per_action="width")
    def write(self) -> float:
        """
        Write energy is zero because transfer costs are already included in the read
        energy.

        Parameters
        ----------
        bits_per_action: int
            The number of bits transferred.

        Returns
        -------
        float
            Energy to transfer data in Joules
        """
        return 0

    @actionDynamicEnergy(bits_per_action="width")
    def transfer(self) -> float:
        """
        Returns the energy to transfer data in Joules.

        Parameters
        ----------
        bits_per_action: int
            The number of bits transferred.

        Returns
        -------
        float
            Energy to transfer data in Joules
        """
        return 20.74e-12


# Original CSV contents:
# tech_node,global_cycle_period,resolution,energy,area,n_instances,action
# 32nm,1e-9,8,1.666666667,1200,1,convert|read
# 32nm,1e-9,8,0,1200,1,leak|update|write
# # Energy: 16*10^-3 W / (1.2*8*10^9 ADC BW) * 10 ^ 12 J->pJ
# # 16*10^-3 / (1.2*8*10^9) * 10 ^ 12
# # Area: 9600um^2 / 8
# # L. Kull et al.," ""A 3.1mW 8b 1.2GS/s single-channel asynchronous SAR ADC
# # with alternate comparators for enhanced speed in 32nm digital SOI
# # CMOS",2013,pp. 468-469,doi: 10.1109/ISSCC.2013.6487818.," 2013 IEEE
# # International Solid-State Circuits Conference Digest of Technical Papers
# # Below are scaled versions based on M. Saberi, R. Lotfi, K. Mafinezhad, W.
# # Serdijn et al., “Analysis of Power Consumption and Linearity in Capacitive
# # Digital-to-Analog Converters used in Successive Approximation ADCs,” 2011.
# # 32nm,1e-9,4,0.79,361.04,1,convert|read
# # 32nm,1e-9,5,0.99,476.91,1,convert|read
# # 32nm,1e-9,6,1.20,626.91,1,convert|read
# # 32nm,1e-9,7,1.42,845.18,1,convert|read
# # 32nm,1e-9,8,1.67,1200,1,convert|read
# # 32nm,1e-9,9,1.969078145,1827.911647,1,convert|read
# # 32nm,1e-9,10,2.379022742,3002.008032,1,convert|read
class IsaacADC(LibraryEstimatorClassBase):
    """
    The analog-digital-converter (ADC) from the ISAAC paper.

    Parameters
    ----------
    tech_node: float
        Technology node in meters.
    resolution: int
        Resolution of the ADC in bits.
    """
    component_name = "isaac_adc"
    priority = 0.9

    def __init__(self, tech_node: float, resolution: int = 8):
        super().__init__(leak_power=0.0, area=1200.0e-12)
        self.tech_node: float = self.scale(
            "tech_node",
            tech_node,
            32e-9,
            tech_node_energy,
            tech_node_area,
            tech_node_leak,
        )
        self.resolution: int = self.scale(
            "resolution", resolution, 8, pow_base(2), pow_base(2), pow_base(2)
        )

    @actionDynamicEnergy
    def convert(self) -> float:
        """
        Returns the energy of one ADC conversion in Joules

        Returns
        -------
        float
            Energy of one ADC conversion in Joules
        """
        return 1.666666667e-12

    @actionDynamicEnergy
    def read(self) -> float:
        """
        Returns the energy of one ADC conversion in Joules

        Returns
        -------
        float
            Energy of one ADC conversion in Joules
        """
        return 1.666666667e-12


# Original CSV contents:
# tech_node,global_cycle_period,width|datawidth,energy,area,action
# 32nm,1e-9,256,20.74,150000,read,energy in pJ;  area in um^2;
# 32nm,1e-9,256,0,150000,leak|update|write
# # ISAAC shares each of these between 4 tiles
# # Assuming router BW = eDRAM BW per tile
# # Power * Time / (Reads+Writes) = Energy per read/write
# # (42e-3 / 4 / 12) power
# # (16384 / ((128*8*10^7*1.2) * 100 / 128)) time for DACs/ADCs to consume entire input buffer
# # (16384 + 2048) / 256 reads+writes
# # (42e-3 / 4 / 12) * (16384 / ((128*8*10^7*1.2) * 100 / 128)) / ((16384 + 2048) / 256) * 1e12
class IsaacRouter(LibraryEstimatorClassBase):
    """
    The router from the ISAAC paper. This is the router shared by four tiles in the
    paper, so divide this area by four if you'd like to get the per-tile area from the
    paper.

    Parameters
    ----------
    tech_node: float
        Technology node in meters.
    width: int
        Width of the router in bits. This is the width of a read/write port.
    """
    component_name = "isaac_router"
    priority = 0.9

    def __init__(self, tech_node: float, width: int = 256):
        super().__init__(leak_power=0.0, area=150000.0e-12)
        self.tech_node: float = self.scale(
            "tech_node",
            tech_node,
            32e-9,
            tech_node_energy,
            tech_node_area,
            tech_node_leak,
        )
        self.width: int = self.scale("width", width, 256, linear, linear, linear)

    @actionDynamicEnergy(bits_per_action="width")
    def read(self) -> float:
        """
        Returns the energy to transfer data in Joules.

        Returns
        -------
        float
            Energy to transfer data in Joules
        """
        return self.transfer()

    @actionDynamicEnergy(bits_per_action="width")
    def write(self) -> float:
        """
        Returns the energy to transfer data in Joules.

        Returns
        -------
        float
            Energy to transfer data in Joules
        """
        return 0


    @actionDynamicEnergy(bits_per_action="width")
    def transfer(self) -> float:
        """
        Returns the energy to transfer data in Joules.

        Parameters
        ----------
        bits_per_action: int
            The number of bits transferred.

        Returns
        """
        return 20.74e-12


# Original CSV contents:
# tech_node,global_cycle_period,width|datawidth,energy,area,action
# 32nm,1e-9,16,0.021,60,shift_add|read|write,energy in pJ;  area in um^2
# 32nm,1e-9,16,0.00E+00,60,leak|update
# # Energy: 16*10^-3 W / (1.2*8*10^9 ADC BW) * 10 ^ 12 J->pJ
# # Energy: .2e-3 W / (1.2*8*10^9 ADC BW) * 10 ^ 12 J->pJ
# # .2e-3 / (1.2*8*10^9) * 10 ^ 12
# # There are 4 of these in an ISAAC IMA
class IsaacShiftAdd(LibraryEstimatorClassBase):
    """
    The shift-and-add unit from the ISAAC paper. This unit will sum and accumulate
    values in a register, while also shifting the register contents to accept various
    power-of-two scaling factors for the summed values.

    Parameters
    ----------
    tech_node: float
        Technology node in meters.
    width: int
        Width of the shift-and-add unit in bits. This is the number of bits of each
        input value that is added to the register.
    """
    component_name = "isaac_shift_add"
    priority = 0.9

    def __init__(self, tech_node: float, width: int = 16):
        super().__init__(leak_power=0.0, area=60.0e-12)
        self.tech_node: float = self.scale(
            "tech_node",
            tech_node,
            32e-9,
            tech_node_energy,
            tech_node_area,
            tech_node_leak,
        )
        self.width: int = self.scale("width", width, 16, linear, linear, linear)

    @actionDynamicEnergy
    def shift_add(self) -> float:
        """
        Returns the energy of one shift-and-add operation in Joules.

        Returns
        -------
        float
            Energy of one shift-and-add operation in Joules
        """
        return 0.021e-12

    @actionDynamicEnergy
    def read(self) -> float:
        """
        Returns the energy to read the shift-and-add unit's output in Joules.

        Returns
        -------
        float
            Energy to read the shift-and-add unit's output in Joules
        """
        return 0

    @actionDynamicEnergy
    def write(self) -> float:
        """
        Returns the energy of one shift-and-add operation in Joules.

        Returns
        -------
        float
            Energy of one shift-and-add operation in Joules
        """
        return 0.021e-12


# Original CSV contents:
# tech_node,global_cycle_period,width|datawidth,energy,area,action
# 32nm,1e-9,1,0.054,29.296875,read,energy in pJ;  area in um^2;
# 32nm,1e-9,1,0,29.296875,leak|update|write,energy in pJ;  area in um^2;
# # Power * Time / (Reads+Writes) = Energy per read/write
# # (7e-3 / 12 W/IMA) power
# # (16384 / ((128*8*10^7*1.2) * 100 / 128)) time for DACs/ADCs to consume entire input buffer
# # (16384 + 2048) * reads+writes
# # (7e-3 / 12) * (16384 / ((128*8*10^7*1.2) * 100 / 128)) / ((16384 + 2048)) * 1e12
# # Assuming bus BW = eDRAM BW
# # Area reported per IMA. In ISAAC, a bus connects 12 IMAs
# # Area: 7500 / (Width 256) = 29.296875 um^2 per bit width
class IsaacEDRAMBus(LibraryEstimatorClassBase):
    """
    The eDRAM bus from the ISAAC paper. This bus connects the eDRAM to the router.

    Parameters
    ----------
    tech_node: float
        Technology node in meters.
    width: int
        Width of the eDRAM bus in bits. This is the width of a read/write port.
    """
    component_name = "isaac_eDRAM_bus"
    priority = 0.9

    def __init__(self, tech_node: float, width: int = 1):
        super().__init__(leak_power=0.0, area=29.296875e-12)
        self.tech_node: float = self.scale(
            "tech_node",
            tech_node,
            32e-9,
            tech_node_energy,
            tech_node_area,
            tech_node_leak,
        )
        self.width: int = self.scale("width", width, 1, linear, linear, linear)

    @actionDynamicEnergy(bits_per_action="width")
    def read(self) -> float:
        """
        Returns the energy to read the eDRAM bus in Joules.

        Parameters
        ----------
        bits_per_action: int
            The number of bits transferred.

        Returns
        -------
        float
            Energy to read the eDRAM bus in Joules
        """
        return self.transfer()

    @actionDynamicEnergy(bits_per_action="width")
    def transfer(self) -> float:
        """
        Returns the energy to transfer data in Joules.

        Parameters
        ----------
        bits_per_action: int
            The number of bits transferred.

        Returns
        -------
        float
            Energy to transfer data in Joules
        """
        return 0.054e-12

    @actionDynamicEnergy(bits_per_action="width")
    def write(self) -> float:
        """
        Returns 0 because transfer costs are already included in the read energy.

        Parameters
        ----------
        bits_per_action: int
            The number of bits transferred.

        Returns
        -------
        float
            Zero
        """
        return 0


# Original CSV contents:
# tech_node,global_cycle_period,resolution,energy,area,rows,action
# 32nm,1e-9,1,0.41667,0.166015625,1,drive|read
# 32nm,1e-9,1,0,0.166015625,1,write|leak|update
# # Energy: 4*10^-3 W / (128*8*10^7*1.2 DAC BW) * 10 ^ 12 J->pJ * 128/100 underutilized due to ADC
# # 4e-3 / (128 * 8 * 1.2 * 10 ^ 7) * 10 ^ 12 * 128/100
# # 0.3255 * 8 * 128 * 1.2e9 / 100 * 1e-9
# # Area: 170um^2 / 128 / 8
class IsaacDAC(LibraryEstimatorClassBase):
    """
    The digital-analog converter (DAC) from the ISAAC paper.

    Parameters
    ----------
    tech_node: float
        Technology node in meters.
    resolution: int
        Resolution of the DAC in bits.
    """
    component_name = "isaac_dac"
    priority = 0.9

    def __init__(self, tech_node: float, resolution: int = 1, rows: int = 1):
        super().__init__(leak_power=0.0, area=0.166015625e-12)
        self.tech_node: float = self.scale(
            "tech_node",
            tech_node,
            32e-9,
            tech_node_energy,
            tech_node_area,
            tech_node_leak,
        )
        self.resolution: int = self.scale(
            "resolution", resolution, 1, pow_base(2), pow_base(2), pow_base(2)
        )
        self.rows: int = self.scale("rows", rows, 1, linear, noscale, noscale)

    @actionDynamicEnergy
    def convert(self) -> float:
        """
        Returns the energy to convert with the the DAC in Joules.

        Returns
        -------
        float
            Energy to convert with the the DAC in Joules
        """
        return 0.41667e-12

    @actionDynamicEnergy
    def read(self) -> float:
        """
        Returns the energy to read the DAC in Joules.

        Returns
        -------
        float
            Energy to read the DAC in Joules
        """
        return 0.41667e-12
